#include "tabbar.h"
#include <settingsprivate.h>
#include "playlist.h"

#include <QApplication>
#include <QIcon>
#include <QMimeData>

#include <QtDebug>

TabBar::TabBar(TabPlaylist *parent) :
	QTabBar(parent), lineEdit(new QLineEdit(this)), tabPlaylist(parent)
{
	this->setAcceptDrops(true);
	this->setDocumentMode(true);
	this->setTabsClosable(true);
	this->setUsesScrollButtons(true);
	this->installEventFilter(this);

	lineEdit->setVisible(false);
	lineEdit->setAlignment(Qt::AlignCenter);
	lineEdit->setFrame(false);
	lineEdit->installEventFilter(this);
	lineEdit->setAttribute(Qt::WA_MacShowFocusRect, false);

	_timer = new QTimer(this);
	_timer->setInterval(300);
	_timer->setSingleShot(true);

	connect(lineEdit, &QLineEdit::returnPressed, this, &TabBar::renameTab);
	// Switch between tabs if the Drag & Drop isn't finished
	connect(_timer, &QTimer::timeout, [=]() {
		this->setCurrentIndex(tabAt(_targetRect.center()));
		_targetRect = tabRect(currentIndex());
	});

	SettingsPrivate *settings = SettingsPrivate::instance();
	auto applyFont = [this, settings] (SettingsPrivate::FontFamily ff, const QFont &newFont)
	{
		if (ff == SettingsPrivate::FF_Playlist) {
			QFont font = newFont;
			font.setPointSizeF(font.pointSizeF() * 0.8);
			this->setFont(font);
			int h = fontMetrics().height() * 1.25;
			if (h >= 30) {
				this->setMinimumHeight(h);
				this->setMaximumHeight(h);
			} else {
				this->setMinimumHeight(30);
				this->setMaximumHeight(30);
			}
		}
	};
	// Init font from settings
	applyFont(SettingsPrivate::FF_Playlist, settings->font(SettingsPrivate::FF_Playlist));
	connect(settings, &SettingsPrivate::fontHasChanged, this, applyFont);
}

/** Trigger a double click to rename a tab. */
void TabBar::editTab(int indexTab)
{
	QMouseEvent dbClickEvent(QEvent::MouseButtonDblClick, tabRect(indexTab).center(), Qt::LeftButton, 0, 0);
	QApplication::sendEvent(this, &dbClickEvent);
}

/** Redefined to accept D&D from another playlist or the library. */
void TabBar::dragEnterEvent(QDragEnterEvent *event)
{
	int idx = tabAt(event->pos());
	if (idx < count() - 1) {
		_targetRect = tabRect(idx);
	} else {
		_targetRect = QRect();
	}
	event->source() ? event->acceptProposedAction() : event->ignore();
}

/** Redefined to accept D&D from another playlist or the library. */
void TabBar::dragMoveEvent(QDragMoveEvent *event)
{
	int idx = tabAt(event->pos());
	if (idx < count()) {
		_targetRect = tabRect(idx);
	}
	Playlist *playlist = qobject_cast<Playlist*>(event->source());
	if (playlist && idx < count()) {
		_timer->start();
	}
	event->source() ? event->acceptProposedAction() : event->ignore();
}

/** Redefined to accept D&D from another playlist or the library. */
void TabBar::dropEvent(QDropEvent *event)
{
	int tab = this->tabAt(event->pos());
	if (Playlist *origin = qobject_cast<Playlist*>(event->source())) {
		Playlist *target = tabPlaylist->playlist(tab);

		// Copy tracks in the target
		QList<QMediaContent> medias;
		for (QModelIndex index : origin->selectionModel()->selectedRows()) {
			medias.append(origin->mediaPlaylist()->media(index.row()));
		}
		// Append tracks at the end
		target->insertMedias(target->model()->rowCount(), medias);

		// Remove tracks from the current playlist if necessary
		if (!SettingsPrivate::instance()->copyTracksFromPlaylist()) {
			origin->removeSelectedTracks();
		}
	} else {
		QByteArray byteArray = event->mimeData()->data("treeview/x-treeview-item");
		if (!byteArray.isEmpty()) {
			QList<QUrl> medias;
			QList<QByteArray> encodedUrls = byteArray.split('|');
			for (QByteArray encodedUrl : encodedUrls) {
				medias << QUrl::fromEncoded(encodedUrl);
			}
			tabPlaylist->insertItemsToPlaylist(-1, medias);
		}
	}

}

/** Redefined to validate new tab name if the focus is lost. */
bool TabBar::eventFilter(QObject *obj, QEvent *event)
{
	// Accept the escape key
	if (obj == lineEdit && event->type() == QEvent::KeyPress) {
		QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
		if (keyEvent->key() == Qt::Key_Escape) {
			lineEdit->releaseMouse();
			lineEdit->setText(lineEdit->text().trimmed());
			lineEdit->close();
		}
	} else if (obj == lineEdit && event->type() == QEvent::FocusOut) {
		lineEdit->releaseMouse();
		lineEdit->setText(lineEdit->text().trimmed());
		renameTab();
	} else if (obj == lineEdit && event->type() == QEvent::MouseButtonPress) {
		QMouseEvent *me = static_cast<QMouseEvent*>(event);
		if (!lineEdit->rect().contains(me->pos())) {
			lineEdit->releaseMouse();
			lineEdit->setText(lineEdit->text().trimmed());
			lineEdit->close();
		}
	} else if (obj == this && event->type() == QEvent::HoverLeave) {
		this->update();
	}
	return false;
}

/** Redefined to display an editable area. */
void TabBar::mouseDoubleClickEvent(QMouseEvent *event)
{
	int tabIndex = tabAt(event->pos());
	int c = currentIndex();
	if (-1 < tabIndex && tabIndex < count() && c == tabIndex) {
		QRect visualRect = tabRect(tabIndex);
		SettingsPrivate *settings = SettingsPrivate::instance();
		if (settings->isRectTabs()) {
			visualRect.setLeft(visualRect.left() + 1);
			visualRect.setRight(visualRect.right() - 1);
		} else {
			visualRect.setLeft(visualRect.left() + 3 + settings->tabsOverlappingLength() * 1.25 + height() / 2);
			visualRect.setRight(visualRect.right() - settings->tabsOverlappingLength());
		}
		visualRect.setTop(visualRect.top() + 1);

		// Disable close buttons in case of unfortunate click
		for (int t = 0; t < count(); t++) {
			QWidget *button = tabButton(t, QTabBar::RightSide);
			if (button && t != tabIndex) {
				button->setEnabled(false);
			}
		}

		// Feed the QLineEdit with current tab text
		lineEdit->setText(tabText(tabIndex));
		lineEdit->selectAll();
		lineEdit->setFocus();
		lineEdit->setGeometry(visualRect);
		lineEdit->setVisible(true);
		lineEdit->grabMouse();
	}
	QTabBar::mouseDoubleClickEvent(event);
}

/** Redefined to validate new tab name without pressing return. */
void TabBar::mousePressEvent(QMouseEvent *event)
{
	int tabIndex = tabAt(event->pos());
	if (lineEdit->isVisible() && ((currentIndex() != tabIndex && lineEdit->isVisible()) || !lineEdit->geometry().contains(event->pos()))) {
		lineEdit->close();
	} else {
		QTabBar::mousePressEvent(event);
	}
}

void TabBar::paintEvent(QPaintEvent *)
{
	SettingsPrivate *settings = SettingsPrivate::instance();
	QStylePainter p(this);
	int dist = 0;
	if (settings->isRectTabs()) {
		paintRectTabs(p);
	} else {
		paintRoundedTabs(p);
	}

	// Global bottom frame border
	int h = tabRect(0).height() - 1;
	p.setPen(QApplication::palette().mid().color());
	if (count() > 1) {
		if (currentIndex() == 0) {
			p.drawLine(tabRect(0).right() + 1 + dist, h, rect().right(), h);
		} else {
			p.drawLine(rect().left(), h, tabRect(currentIndex()).left(), h);
			p.drawLine(tabRect(currentIndex()).right() + 1 + dist, h, rect().right(), h);
		}
	} else {
		if (isLeftToRight()) {
			p.drawLine(tabRect(0).width() + dist, h, rect().right(), h);
		} else {
			p.drawLine(rect().left(), h, tabRect(0).left(), h);
		}
	}
}

/** Redefined to return a square for the last tab which is the [+] button. */
QSize TabBar::tabSizeHint(int index) const
{
	QSize s = QTabBar::tabSizeHint(index);
	if (!SettingsPrivate::instance()->isRectTabs()) {
		s.setWidth(s.width() + SettingsPrivate::instance()->tabsOverlappingLength() * 2);
	}
	// Adjust height to the minimum otherwise a small gap might appear between tab and header (depending of font size)
	if (s.height() > height()) {
		s.setHeight(height());
	}
	return s;
}

void TabBar::paintRectTabs(QStylePainter &p)
{
	static const qreal penScaleFactor = 0.2;
	for (int i = 0; i < count(); i++) {
		QStyleOptionTab o;
		initStyleOption(&o, i);

		// Background color
		p.save();
		if (i != currentIndex()) {
			o.rect.adjust(0, 2, 0, 0);
		} else if (i == count()) {
			o.rect.adjust(2, 2, -4, -4);
		}

		// Highlight the tab under the cursor
		if (o.state.testFlag(QStyle::State_MouseOver) && i != currentIndex()) {
			p.setPen(QPen(o.palette.highlight(), penScaleFactor));
			p.fillRect(o.rect, o.palette.highlight().color().lighter());
		} else {
			p.setPen(QPen(o.palette.mid(), penScaleFactor));
			if (i == currentIndex()) {
				/// XXX
				if (SettingsPrivate::instance()->isCustomColors()) {
					p.fillRect(o.rect, o.palette.base().color().lighter(110));
				} else {
					p.fillRect(o.rect, o.palette.base());
				}
			} else {
				p.fillRect(o.rect, o.palette.window());
			}
		}

		if (o.state.testFlag(QStyle::State_MouseOver)) {
			p.setPen(o.palette.highlight().color());
		} else {
			p.setPen(o.palette.mid().color());
		}
		// Frame tab, it is not a rectangle but only 3 lines
		p.drawLine(o.rect.x(), o.rect.y(), o.rect.x(), o.rect.y() + o.rect.height());
		p.drawLine(o.rect.x() + o.rect.width(), o.rect.y(), o.rect.x() + o.rect.width(), o.rect.y() + o.rect.height());
		p.drawLine(o.rect.x(), o.rect.y(), o.rect.x() + o.rect.width(), o.rect.y());
		//}
		p.restore();

		// Icon
		QRect r = tabRect(i);
		r.setHeight(fontMetrics().ascent());
		r.translate(10, (height() - r.height()) / 2);
		r.setWidth(r.height() / 2);
		p.setRenderHint(QPainter::SmoothPixmapTransform);
		o.icon.paint(&p, r, Qt::AlignLeft | Qt::AlignVCenter);

		// Playlist name
		if (i == currentIndex()) {
			p.setPen(o.palette.text().color());
		} else if (o.state.testFlag(QStyle::State_MouseOver)) {
			if (SettingsPrivate::instance()->isCustomTextColorOverriden()) {
				p.setPen(o.palette.highlightedText().color());
			} else {
				p.setPen(o.palette.text().color());
			}
		} else {
			p.setPen(o.palette.mid().color());
		}
		QRect rText(r.x() + r.width() + 10, r.y(), o.rect.width() - r.width() - 10, r.height());
		p.drawText(rText, Qt::AlignLeft | Qt::AlignVCenter, o.text);
	}
}

void TabBar::paintRoundedTabs(QStylePainter &p)
{
	/// TODO: minor highlight bug when mouse goes on another tab without click
	// Draw all tabs before the selected tab
	QList<int> tabs;
	for (int i = 0; i < count(); i++)
		if (currentIndex() != i) tabs.append(i);
	tabs.append(currentIndex());

	int dist = SettingsPrivate::instance()->tabsOverlappingLength();
	for (int idx = 0; idx < count(); idx++) {
		int i = tabs.at(idx);
		QStyleOptionTab o;
		initStyleOption(&o, i);

		// Background color
		if (i != currentIndex()) {
			o.rect.adjust(0, 2, 0, 0);
		} else if (i == count()) {
			o.rect.adjust(2, 2, -4, 0);
		}

		/// Adjust parameters to tighten tabs
		//o.rect.adjust(-dist / 2, 0, dist / 2, 0);

		// Rounded frame tab
		QPainterPath ppLeft;
		ppLeft.moveTo(o.rect.x() + dist * 0, o.rect.y() + o.rect.height() - 1);
		ppLeft.cubicTo(o.rect.x() + dist * 1, o.rect.y() + o.rect.height() - 1,
					   o.rect.x() + dist * 1, o.rect.y(),
					   o.rect.x() + dist * 2, o.rect.y());

		ppLeft.lineTo(o.rect.x() + o.rect.width() - dist * 1, o.rect.y());
		ppLeft.cubicTo(o.rect.x() + o.rect.width() - dist * 0, o.rect.y(),
						o.rect.x() + o.rect.width() - dist * 0, o.rect.y() + o.rect.height() - 1,
						o.rect.x() + o.rect.width() + dist * 1, o.rect.y() + o.rect.height() - 1);

		p.save();
		if (o.state.testFlag(QStyle::State_MouseOver)) {
			p.setPen(o.palette.highlight().color());
		} else {
			p.setPen(o.palette.mid().color());
		}
		if (i == currentIndex()) {
			p.fillPath(ppLeft, o.palette.base());
		} else if (o.state.testFlag(QStyle::State_MouseOver)) {
			p.fillPath(ppLeft, o.palette.highlight().color().lighter());
		} else {
			p.fillPath(ppLeft, o.palette.window());
		}
		p.setRenderHint(QPainter::Antialiasing, true);
		p.drawPath(ppLeft);
		p.setRenderHint(QPainter::Antialiasing, false);

		p.restore();

		/// DEBUG
		//p.drawRect(o.rect);

		// Playlist name
		if (i == currentIndex()) {
			p.setPen(o.palette.text().color());
		} else if (o.state.testFlag(QStyle::State_MouseOver)) {
			if (SettingsPrivate::instance()->isCustomTextColorOverriden()) {
				p.setPen(o.palette.highlightedText().color());
			} else {
				p.setPen(o.palette.text().color());
			}
		} else {
			p.setPen(o.palette.mid().color());
		}

		int pm = style()->pixelMetric(QStyle::PM_TabCloseIndicatorWidth, &o, this);
		QRect rText(o.rect.x() + pm * 2, o.rect.y(), o.rect.width() - 2 * pm, o.rect.height());
		//qDebug() << Q_FUNC_INFO << i << rText;
		p.drawText(rText, Qt::AlignLeft | Qt::AlignVCenter, o.text);

		// Icon
		//QRect r = tabRect(i);
		//r.setHeight(fontMetrics().ascent());
		//r.translate(dist, (height() - r.height()) / 2);
		//r.setWidth(r.height() / 2);
		p.setRenderHint(QPainter::SmoothPixmapTransform);

		int vpm = style()->pixelMetric(QStyle::PM_TabCloseIndicatorHeight);
		QRect iconRect = style()->subElementRect(QStyle::SE_CheckBoxContents, &o, this);
		//qDebug() << Q_FUNC_INFO << iconRect;
		iconRect.translate(rText.right(), 0);
		int w = style()->pixelMetric(QStyle::PM_TabBarIconSize);
		iconRect.setWidth(w);
		//iconRect.setX(rText.right());
		iconRect.setY(iconRect.y() + (iconRect.height() - vpm) / 2);
		iconRect.setHeight(vpm);

		QRect r2(o.rect.x() + o.rect.width() - w - dist / 2, o.rect.y() + (o.rect.height() - w) / 2,
				 w, w);

		//qDebug() << Q_FUNC_INFO << i << o.state;
		//qDebug() << Q_FUNC_INFO << i << o.state.testFlag(QStyle::State_Active) << o.state.testFlag(QStyle::State_On);

		o.icon.paint(&p, r2, Qt::AlignLeft | Qt::AlignVCenter);
		//p.drawPixmap(r2, QPixmap(":/icons/playlistIcon"));
		//o.rect.setRight(iconRect.right());
	}
}

/** Rename a tab. */
void TabBar::renameTab()
{
	this->setTabText(currentIndex(), lineEdit->text());
	emit tabRenamed(currentIndex(), tabText(currentIndex()));
	lineEdit->close();
	for (int t = 0; t < count(); t++) {
		QWidget *button = tabButton(t, QTabBar::RightSide);
		if (button) {
			button->setEnabled(true);
		}
	}
}
